#load "pa_extend.cmo";;
#load "q_MLast.cmo";;

open Pcaml
open MLast

let loc = Ploc.dummy

module Utils =
struct
  let compose f g = fun v -> f (g v)
  let id = fun x -> x
  let contains_aux equal ls item = List.exists (equal item) ls
  let contains ls item = contains_aux (=) ls item
  (*remove duplicates*)
  let unique_filter equal ls = 
    List.fold_right 
      (fun item res -> res @ if contains_aux equal res item then [] else [item])
      ls
      []
  (*split list of triples*)
  let split3 ls = List.fold_left 
    (fun (ls1, ls2, ls3) (e1, e2, e3) -> (e1::ls1, e2::ls2, e3::ls3))
    ([],[],[])
    ls
  (*lists product*)
  let prod_lists lls = 
    List.fold_right 
      (fun ls res -> 
        List.concat (List.map (fun el -> List.map (fun l -> el :: l) res) ls))
      lls 
      [[]]
end

module Name = 
struct
  let marker t = t ^ "_marker"
  let typeinfo t = t
  let traverse t = t ^ "_traverse"
  let result t = t ^ "_result"
  let param = typeinfo
  let constructor_marker = Utils.compose marker String.uppercase
  let accumulator t = t ^ "_accumulator"
  let specialization = "spec"
  let arg_prefix = "traverse_argument"
  let args_names_generator () =
    let name = ref 0 in
    fun () -> arg_prefix ^ (string_of_int (name := !name + 1; !name))
end

module AST_utils = 
struct
  (*convert expession list to expression that represents list*)
  let list_to_expr ls =
    List.fold_left
      (fun res next -> <:expr< [$next$ :: $res$] >>)
      <:expr<[]>>
      (List.rev ls)
  (*create expression that concatenates lists*)
  let concat ls = <:expr<List.concat $list_to_expr ls$>>
end

module Type_utils =
struct
  let name_by_decl td =
    let Ploc.VaVal (_, Ploc.VaVal name) = td.tdNam in
    name
  let ast_by_decl td = td.tdDef
  let params_by_decl td =
    List.map 
      (function (Ploc.VaVal (Some name), _) -> name)
      (match td.tdPrm with Ploc.VaVal params -> params)
  (*module that stores type constructor*)
  let tycon con_name param td = 
    let td = { tdNam = (Ploc.VaVal (loc, Ploc.VaVal "tc"));
               tdPrm = (Ploc.VaVal [Ploc.VaVal (Some param), None]);
               tdDef = (<:ctyp<$lid:con_name$ '$param$>>);
               tdPrv = td.tdPrv;
               tdCon = td.tdCon }
    in
    let tdi = <:str_item<type $list:[td]$>> in
    <:module_expr<struct $list:[tdi]$ end>>
end

module Type_atom =
struct
  type acc = string list * string
  and atom = Name of acc | App of acc * atom | Var of string

  let rec from_ast ast =
    match ast with
    | <:ctyp< $uid:m$ . $t$ >> -> 
      (match from_ast t with
      | Some (Name (ms, t)) -> Some (Name (m :: ms, t))
      | _ -> None)
    | <:ctyp< $c$ $p$ >> ->
      let c_atom = from_ast c in
      let p_atom = from_ast p in
      (match c_atom with
      | Some (Name (ms, t)) -> 
        (match p_atom with
        | Some p -> Some (App ((ms, t), p))
        | None -> None)
      | _ -> None)
    | <:ctyp< '$v$ >> -> Some (Var v)
    | <:ctyp< $lid:t$ >> -> Some (Name ([], t))
    | _ -> None

  let rec to_name atom =
    let sep = "_SEP_" in
    let acc_to_string (ls, t) = 
      "accessBEGIN_" 
      ^ (List.fold_left (fun s e -> e ^ sep ^ s) "" ls) 
      ^ sep ^ t ^ "_accessEND"
    in
    match atom with
    | Name (ms, t) -> acc_to_string (ms, t)
    | App ((ms, t), p) ->
      "applicationBEGIN_" 
      ^ acc_to_string (ms, t)
      ^ "_TO_" ^ to_name p
      ^ "_applicationEND"
    | Var s -> "variable_" ^ s

  let rec to_typeinfo local_types atom =
    let acc_to_typeinfo (ms, t) = 
      List.fold_left 
        (fun f m e -> <:expr< $uid:m$ . $f e$>>) 
        Utils.id 
        ms 
        <:expr<$lid:Name.typeinfo t$>> 
    in
    match atom with 
    | Name ([], t) when Utils.contains local_types t -> 
      <:expr<($lid:Name.typeinfo t$ ())>> 
    | Name (ms, t) -> acc_to_typeinfo (ms, t)
    | App ((ms, t), p) -> 
      <:expr< $acc_to_typeinfo (ms, t)$ $to_typeinfo local_types p$ >>
    | Var s -> <:expr<$lid:Name.param s$>>

  let rec equal a b =
    let acc_equal (ms, t) (ms', t') = 
      t = t' 
      && List.length ms = List.length ms' 
      && List.fold_left2 (fun acc a b -> acc && a = b) true ms ms'
    in
    match a with
    | Name (ms, t) -> 
      (match b with
      | Name (ms', t') -> acc_equal (ms, t) (ms', t')
      | _ -> false)
    | App ((ms, t), p) ->
      (match b with
      | App ((ms', t'), p') -> acc_equal (ms, t) (ms', t') && equal p p'
      | _ -> false)
    | Var s -> 
      (match b with
      | Var s' -> s = s'
      | _ -> false)

  let rec collect ast =
    let check_atom ast =
      match from_ast ast with
      | Some atom -> [atom]
      | None -> failwith "this type expression is not supported"
    in
    match ast with
    | <:ctyp< $t1$ . $t2$ >> -> check_atom ast
    | <:ctyp< $t1$ $t2$ >> -> check_atom ast
    | <:ctyp< $lid:t1$ >> -> check_atom ast
    | <:ctyp< ' $t1$ >> -> check_atom ast      
    | <:ctyp< [ $list:llslt$ ] >> ->
      let sub_types = List.map
        (fun l -> let _, Ploc.VaVal name, Ploc.VaVal lt, _ = l in lt)
        llslt
      in
      let lt = List.fold_left (@) [] sub_types in
      List.fold_left (@) [] (List.map collect lt)
    | <:ctyp< ( $list:lt$ ) >> ->
      List.fold_left (@) [] (List.map collect lt)
    | MLast.TyVrn (_, (Ploc.VaVal lpv), _) ->
      let sub_types = List.map
        (fun pv -> 
          let PvTag (_, Ploc.VaVal name, _, Ploc.VaVal lt) = pv in lt)
        lpv
      in
      let lt = List.concat sub_types in
      List.concat (List.map collect lt)
end

let create_typeinfo_decl local_types td =
  let type_name  = Type_utils.name_by_decl td in
  let type_ast   = Type_utils.ast_by_decl  td in
  let type_param =
    match Type_utils.params_by_decl td with
    | [p] -> Some p
    | []  -> None
    | _   -> failwith "multiple parameters are not supported"
  in  
  let marker_decl = match type_param with 
    | None -> 
      <:str_item<value $lid:Name.marker type_name$ : marker $lid:type_name$ = make ()>>
    | Some p -> 
      let tycon = Type_utils.tycon type_name p td in
      <:str_item<module $uid:Name.constructor_marker type_name$ = Make ($tycon$)>>
  in
  let typeinfo =
    let marker_item = match type_param with
      | None -> (<:patt<marker>>, <:expr<$lid:Name.marker type_name$>>)
      | Some p -> (<:patt<marker>>,
                   <:expr<($uid:Name.constructor_marker type_name$ . make)
                          ($lid:Name.param p$ . marker)>>)
    in
    let traverse_func specialize bind constructor =
      let atoms = 
        Utils.unique_filter Type_atom.equal (Type_atom.collect type_ast) 
      in
      let precalculate expr =
        let combine expr atom =
          let typeinfo = Type_atom.to_typeinfo local_types atom in
          let name = Type_atom.to_name atom in
          <:expr<let $lid:Name.traverse name$ = $specialize typeinfo$ in $expr$>>
        in
        <:expr<$List.fold_left combine expr atoms$>>
      in
      let next_name = Name.args_names_generator () in
      let gen_result_func type_name type_ast =
        let rec gen_matching type_ast = 
          let gen_matching_aux variants create_value create_patt subtypes flat = 
            List.concat (List.map
                           (fun var -> 
                             let sub_res = Utils.prod_lists 
                               (List.map gen_matching (List.rev (subtypes var))) in
                             List.map 
                               (fun s ->
                                 let patts, funcs, results = Utils.split3 s in
                                 (create_patt var patts, 
                                  List.fold_left Utils.compose Utils.id funcs,
                                  constructor (create_value var) flat results))
                               sub_res)
                           variants)
          in
          let by_name tn =
            let name : string = next_name () in
              [(<:patt< $lid:name$ >>, 
                (fun e -> <:expr<$lid:Name.traverse tn$  
                  (fun $lid:Name.accumulator name$ -> $bind (Name.accumulator name) e$)
                  $lid:name$>>),
                constructor 
                  (function e::[] -> e)
                  AST_utils.list_to_expr
                  [<:expr<$lid:Name.accumulator name$>>])]
          in
          let treat_atom ast =
            match Type_atom.from_ast ast with
            | Some atom -> by_name (Type_atom.to_name atom)
            | None -> failwith "this type expression is not supported"
          in
          (match type_ast with
          | <:ctyp< $t1$ . $t2$ >> -> treat_atom type_ast
          | <:ctyp< $t1$ $t2$ >> -> treat_atom type_ast
          | <:ctyp< $lid:s$ >> -> treat_atom type_ast
          | <:ctyp< ' $s$ >> -> treat_atom type_ast
          | <:ctyp< [ $list:llslt$ ] >> ->
            let variants = List.map 
              (fun l -> let _, Ploc.VaVal n, Ploc.VaVal lt, _ = l in n, lt)
              llslt
            in
            let create_value (name, _) es =
              match es with
              | [] -> <:expr<$uid:name$>>
              | e::[] -> <:expr<$uid:name$ $e$>>
              | _ -> <:expr<$uid:name$ ($list:es$)>>
            in
            let create_patt (name, _) ps =
              match ps with
              | [] -> <:patt<$uid:name$>>
              | p::[] -> <:patt<$uid:name$ $p$>>
              | _ -> <:patt< $uid:name$ ($list:ps$)>>
            in
            gen_matching_aux variants create_value create_patt snd AST_utils.concat
          | <:ctyp< ( $list:lt$ ) >> ->
            let create_value _ es = <:expr< ( $list:es$ ) >> in
            let create_patt _ patts = <:patt< ( $list:patts$ ) >> in     
            gen_matching_aux [lt] create_value create_patt Utils.id AST_utils.concat
          | MLast.TyVrn (loc, (Ploc.VaVal lpv), _) ->
            let variants = List.map
              (fun (PvTag (_, Ploc.VaVal name, _, Ploc.VaVal lt)) -> name, lt)
              lpv
            in
            let create_value (name, _) es =
              match es with 
              | [] -> <:expr< ` $name$>>
              | e::[] -> <:expr< ` $name$ $e$>>
              | _ -> <:expr< ` $name$ ($list:es$)>>
            in
            let create_patt (name, _) ps =
              match ps with 
              | [] -> <:patt< ` $name$>>
              | p::[] -> <:patt< ` $name$ $p$>>
              | _ -> <:patt< ` $name$ ($list:ps$)>> 
            in
            gen_matching_aux variants create_value create_patt snd AST_utils.concat)
        in
        let matching = List.map
          (fun patt, func, expr -> patt, Ploc.VaVal None, func <:expr< cont $expr$>>) 
          (gen_matching type_ast)
        in
        (<:patt<$lid:Name.result type_name$>>,
         <:expr<(fun cont e -> match e with [ $list:matching$ ])>>) (*TODO names*)
      in
      let value = precalculate 
        <:expr<let rec $list:[gen_result_func type_name type_ast]$ in 
               $lid:Name.result type_name$>>
      in <:expr<(fun $lid:Name.specialization$ -> $value$)>>
    in
    (*TODO parent and flatten abstraction refactoring*)
    let gmapT_item =
      let constructor parent flatten exprs = parent exprs in
      let specialize ti = <:expr<$lid:Name.specialization$.transform $ti$>> in
      (<:patt<gmapT>>, traverse_func specialize (fun _ a -> a) constructor)
    in
    let gmapQ_item = 
      let constructor parent flatten exprs = flatten exprs in
      let specialize ti = <:expr<$lid:Name.specialization$.query $ti$>> in
      (<:patt<gmapQ>>, traverse_func specialize (fun _ a -> a) constructor)
    in
    let items = [marker_item; gmapT_item; gmapQ_item] in
    <:expr<{ $list:items$ }>>
  in
  let typeinfo_val = match type_param with
    | None -> <:expr<(fun () -> $typeinfo$)>>
    | Some p -> <:expr<(fun $lid:Name.param p$ -> $typeinfo$)>>
  in
  let typeinfo_decl = (<:patt<$lid:Name.typeinfo type_name$>>, typeinfo_val) in
  (marker_decl, typeinfo_decl)

EXTEND
  GLOBAL: str_item;
  str_item: LEVEL "top" [
    ["datatype" ; tdl = LIST1 type_decl SEP "and"->
    let local_types = 
      List.map 
        Type_utils.name_by_decl 
        (List.filter 
           (fun td -> match Type_utils.params_by_decl td with 
           | [] -> true 
           | _ -> false) 
           tdl) 
    in
    let typeinfo_decls = List.map (create_typeinfo_decl local_types) tdl in
    let marker_decl_items = List.map fst typeinfo_decls in
    let typeinfo_decl_item = 
      <:str_item<value rec $list:List.map snd typeinfo_decls$>>
    in
    let type_decl_item = <:str_item<type $list:tdl$>> in
    let typeinfo_aux_decl = List.map
      (fun name -> 
        <:str_item<value $lid:Name.typeinfo name$ = $lid:Name.typeinfo name$ ()>>)
      local_types
    in
    let decls = 
        [type_decl_item] 
      @ marker_decl_items 
      @ [typeinfo_decl_item] 
      @ typeinfo_aux_decl
    in
    <:str_item<declare $list:decls$ end>> ]
  ];   
  
END
