open Marker
open Generic
open Std_types

(*
  Test data types:
*)

datatype var = string
datatype expr = 
    Add   of expr * expr
  | Neg   of expr
  | Const of int
  | Var   of var

datatype company = C of dept list
and dept         = D of name * manager * subunit list
and subunit      = PU of employee | DU of dept
and employee     = E of person * salary
and person       = P of name * address
and salary       = S of float
and manager      = employee
and name         = string
and address      = string

(*
  Test data generator:
*)

let divide n = 
  let left = (Random.int (n - 1)) + 1  in
  let right = n - left in
  (left, right)

let apollo = Const 0
let dionysus = Var "wine"

let generate_expr seed size =
  Random.init seed;
  let rec conquer remains cont =
    match remains with
    | 1 -> 
      (match (Random.int 2) with
      | 0 -> cont apollo
      | 1 -> cont dionysus)
    | 2 -> cont (Add (conquer 1 (fun x -> x), conquer 1 (fun x -> x)))
    | _ ->
      let left, right = divide (remains - 1) in
      conquer left (fun l -> conquer right (fun r -> cont (Add (l, r))))
  in
  conquer size (fun x -> x)

let zeus = E (P ("Zeus", "Olympus"), S 0.0)
let nero = E (P ("Nero", "Rome"), S 20.0)
let petronius = E (P ("Petronius", "Rome"), S 10.0)

let roman_empire = D ("Roman Empire", nero, [PU petronius])
let create_node l r = 
  D ("Cosmos", zeus, [DU l; DU r])

let generate_company seed size = 
  Random.init seed;
  let rec conquer remains cont =
    match remains with
    | 1 | 2 -> cont roman_empire
    | _ -> 
      let left, right = divide (remains - 1) in
      conquer left (fun l -> conquer right (fun r -> cont (create_node l r)))
  in
  match size with 
  | 1 | 2 -> C [roman_empire]
  | _ -> 
    let l, r = divide (size - 1) in
    C [conquer l (fun x -> x); conquer r (fun x -> x)]

module Aux :
sig
  val map_to_string : ('a -> string) -> 'a list -> string
end =
struct
  let map_to_string f ls =
    let ls = List.map f ls in
    let rec render = function
      | h :: [] -> h
      | h :: t -> h ^ "; " ^ render t
    in
    match ls with
    | [] -> "[]"
    | _ -> "[" ^ render ls ^ "]"
end

(*
  To string:
*)

let rec company_to_string = 
  function (C deps) -> 
    "C (" ^  (Aux.map_to_string dept_to_string deps) ^ ")"
and dept_to_string =
  function (D (name, man, sunits)) ->
    "D (" ^ name ^ ", " ^ 
            employee_to_string man ^ ", " ^ 
            Aux.map_to_string subunit_to_string sunits ^ ")"
and subunit_to_string = function
  | PU emp -> "PU " ^ employee_to_string emp 
  | DU dept -> "DU " ^ dept_to_string dept
and employee_to_string =
  function (E (person, salary)) ->
    "E (" ^ person_to_string person ^ ", " ^ salary_to_string salary ^ ")"
and person_to_string =
  function (P (name, address)) ->
    "P (" ^ name ^ ", " ^ address ^ ")"
and salary_to_string =
  function (S salary) ->
    "S " ^ string_of_float salary

let rec expr_to_string = function
  | Add  (x, y) -> expr_to_string x ^ " + " ^ expr_to_string y 
  | Neg   x     -> "-" ^ expr_to_string x
  | Const x     -> string_of_int x
  | Var   x     -> x
