open Type_eq
open Marker


type 'a typeinfo = { marker : 'a marker; 
                     gmapT : 't . 't transform -> ('a -> 't) -> ('a -> 't);
                     gmapQ : 't . 't query -> ('t list -> 't) -> ('a -> 't) }
and 't transform = { transform : 'a . 'a typeinfo -> ('a -> 't) -> ('a -> 't) }
and 't query  = { query : 'a . 'a typeinfo -> ('t -> 't) -> ('a -> 't) }

module type Monad =
sig
  type 'a t
  val bind : 'a t -> ('a -> 'b t) -> 'b t
  val return : 'a -> 'a t
end


module Lift_t : functor (M : Monad) ->
sig
  type lifted
  val apply : lifted -> 'a marker -> 'a -> 'a M.t
  val return : lifted
  val extend : lifted -> 'a typeinfo -> ('a -> 'a M.t) -> lifted
  val lift : 'a typeinfo -> ('a -> 'a M.t) -> lifted
end = functor (M : Monad) ->
struct
  type lifted = { f : 'a . 'a marker -> 'a -> 'a M.t }
  let apply f = fun m v -> f.f m v  
  let return = { f = fun _ -> M.return }

  let extend source a func =
    { f = fun b ->
      let a = a.marker in
      match compare a b with
      | None   -> source.f b
      | Some e -> 
        let forward = cast (symm e) in
        let module S = Subst (struct type 'a tc = 'a M.t end) in
        let backward = cast (S.subst e) in
        (fun v -> backward (func (forward v))) }

  let lift a func = extend return a func
end

module Lift :
sig
  type lifted
  val apply : lifted -> 'a marker -> 'a -> 'a
  val return : lifted
  val extend : lifted -> 'a typeinfo -> ('a -> 'a) -> lifted
  val lift : 'a typeinfo -> ('a -> 'a) -> lifted
  val is_interesting : 'a marker -> lifted -> bool
end =
struct
  type lifted_aux = { f : 'a . 'a marker -> 'a -> 'a }
  type lifted = lifted_aux * untyped_marker list
  let apply (f, _) m = f.f m
  let return = { f = fun _ -> (fun x -> x) }, []

  let extend (source, source_markers) a func =
    let a = a.marker in
    { f = fun b ->
      match compare a b with
      | None   -> source.f b
      | Some e -> 
        let forward = cast (symm e) in
        let backward = cast e in
        (fun v -> backward (func (forward v))) }, (untype a) :: source_markers

  let lift a func = extend return a func

  let is_interesting m (_, ms) = 
    List.exists (equals (untype m)) ms
end

module Lift_q :
sig
  type 'r lifted
  val apply : 'r lifted -> 'a marker -> 'a -> 'r list -> 'r
  val default : ('r list -> 'r) -> 'r lifted
  val extend : 'r lifted -> 'a typeinfo -> ('a -> 'r list -> 'r) -> 'r lifted
  val lift : ('r list -> 'r) -> 'a typeinfo -> ('a -> 'r list -> 'r) -> 'r lifted
end =
struct
  type 'r lifted = { f : 'a . 'a marker -> 'a -> 'r list -> 'r }
  let apply f = fun m v ch -> f.f m v ch
  let default combine = { f = fun _ _ ls -> combine ls }

  let extend source a func =
    { f = fun b ->
      let a = a.marker in
      match compare a b with
      | None   -> source.f b
      | Some e -> (fun v ch -> func (cast (symm e) v) ch) }

  let lift combine a func = extend (default combine) a func
end


(*

everywhere, everything, etc.

*)

module Tht = Hht.Make (struct type ('a, 'b) t = ('b -> 'a) -> ('b -> 'a) end)
module Qht = Hht.Make (struct type ('a, 'b) t = ('a list -> 'a) -> ('b -> 'a) end)

let stub = fun a b -> failwith "Fixed point error"

module Interesting :
sig
  type t
  val create : unit -> t
  val update : t -> untyped_marker -> bool -> t
  val is_interesting : t -> untyped_marker -> bool
  val equals : t -> t -> bool
end = 
struct
  type t = (untyped_marker * bool) list

  let create () = []

  let rec update ls m v =
    match ls with 
    | [] -> [m, v]
    | (h, hv) :: t -> 
      if Marker.equals h m then (m, v) :: t
      else (h, hv) :: (update t m v)

  let rec is_interesting ls m =
    match ls with
    | [] -> false
    | (h, v) :: t -> 
      if Marker.equals h m then v
      else is_interesting t m

  let equals l1 l2 =
    let belongs l1 l2 = 
      List.fold_left (fun acc (h, v) -> acc && (is_interesting l2 h = v)) true l1
    in
     belongs l1 l2 && belongs l2 l1
end

let rec dfs_children m = function
  | [] -> []
  | h :: t -> if equals m h then t else dfs_children m t

module Transform =
struct
  let everywhere ti f =
    let saved = Tht.create () in
    let analysis = ref true in
    let interesting = ref (Interesting.create ()) in
    let dfs_log = ref [] in
    let rec transform : 'a . 'a typeinfo -> ('a -> 't) -> ('a -> 't) = 
      fun ti -> 
        let cur_is_interesting = Lift.is_interesting ti.marker f in
        if !analysis then
          (* analysis *)
          let _ = 
            if List.exists (equals (untype ti.marker)) !dfs_log then
              dfs_log := !dfs_log @ [untype ti.marker]
            else
              (let next_log = !dfs_log @ [untype ti.marker] in
               dfs_log := next_log;
               let _ = ti.gmapT { transform = transform } in
               interesting := Interesting.update !interesting (untype ti.marker)
                 (cur_is_interesting || 
                    (List.fold_left 
                       (fun acc v -> acc || Interesting.is_interesting !interesting v)
                       false
                       (dfs_children (untype ti.marker) (!dfs_log))
                    ));
               dfs_log := next_log)
          in
          stub          
        else
          (* calculation *)
          let f = Lift.apply f ti.marker in
          let children_interesting = 
            Interesting.is_interesting !interesting (untype ti.marker)
          in
          try
            let tr = Tht.find saved ti.marker in
            if cur_is_interesting && children_interesting then
              fun cont v -> !tr (fun v -> cont (f v)) v
            else if not cur_is_interesting && children_interesting then
              fun cont v -> !tr cont v
            else if cur_is_interesting && not children_interesting then
              fun cont v -> cont (f v)
            else fun cont v -> cont v
          with
            _ ->
              let tr = ref stub in
              Tht.add saved ti.marker tr;
              tr := ti.gmapT { transform = transform };
              Tht.remove saved ti.marker;
              if cur_is_interesting && children_interesting then
                fun cont v -> !tr (fun v -> cont (f v)) v
              else if not cur_is_interesting && children_interesting then
                !tr
              else if cur_is_interesting && not children_interesting then
                fun cont v -> cont (f v)
              else fun cont v -> cont v
    in
    let rec fix_interesting ti = 
      let old_int = !interesting in
      let _ = transform ti in
      if not (Interesting.equals !interesting old_int) 
      then fix_interesting ti
    in
    fix_interesting ti;
    analysis := false;
    transform ti
end

module TransformM = functor (M : Monad) ->
struct
  module Lift = Lift_t (M)
  let everywhere ti f =
    let saved = Tht.create () in
    let rec transform : 'a . 'a typeinfo -> ('a -> 't M.t) -> ('a -> 't M.t) = 
      fun ti -> 
        let f v = Lift.apply f ti.marker v in
        try
          let tr = Tht.find saved ti.marker in
          fun cont v -> !tr (fun v -> M.bind (f v) cont) v
        with
            _ ->
              let tr = ref stub in
              Tht.add saved ti.marker tr;
              tr := ti.gmapT { transform = transform };
              Tht.remove saved ti.marker;
              fun cont v -> !tr (fun v -> M.bind (f v) cont) v
    in
    transform ti
end

module Query =
struct
  let everything ti f = 
    let saved = Qht.create () in
    let rec query : 'a . 'a typeinfo -> ('t -> 't) -> ('a -> 't) =
      fun ti -> 
        try
          let q = Qht.find saved ti.marker in
          fun cont v -> !q (fun r -> cont (Lift_q.apply f ti.marker v r)) v
        with
            _ ->
              let q = ref stub in
              Qht.add saved ti.marker q;
              q := ti.gmapQ { query = query };
              Qht.remove saved ti.marker;
              fun cont v -> !q (fun r -> cont (Lift_q.apply f ti.marker v r)) v
    in
    query ti
end
